-- declare spell names and IDs
local vf = GetSpellInfo(194249)
local stm = GetSpellInfo(193223)
local pi = GetSpellInfo(10060)
local disp = GetSpellInfo(47585)
local sa = 148859
local swp = 589
local mb = 8092
local vb = 205448
local swd = 32379
local swv = 205351
local sc = 205385

-- declare bar related stuff
local mainFrame = CreateFrame("Frame", nil, UIParent)
local bgFrame = CreateFrame("Frame", "InsanityBGFrame", mainFrame)
local barFrame = CreateFrame("Frame", "InsanityBarFrame", bgFrame)
local drainFrame = CreateFrame("Frame", "InsanityDrainFrame", barFrame)
local capFrame = CreateFrame("Frame", "InsanityCapFrame", bgFrame)
local threshFrame = CreateFrame("Frame", "InsanityThreshFrame", bgFrame)
local stackFrame = CreateFrame("Frame", "InsanityStackFrame", bgFrame)
local border = 3
local soundPath = "Sound/Doodad/FX_SONIC_CAST_10.OGG" --"Sound/SPELLS/shadowfury_impact_base.ogg"
local defaultSettings = {width=500, height=20, showExtra=true, showTimes=true, showStacks=true, hideOutsideVF=true, hideOutsideCombat=false, lock=false, playSound=false}
local barWidth = defaultSettings.width
local barHeight = defaultSettings.height

-- declare settings panel stuff
local addon = {}
addon.panel = CreateFrame("Frame", "InsanityBarPanel", UIParent)
addon.panel.name = "InsanityBar"
addon.loaded = false
local knobs = {}
knobs.width = CreateFrame("Slider", "InsanityBarWidthScale", addon.panel, "OptionsSliderTemplate")
knobs.height = CreateFrame("Slider", "InsanityBarHeightScale", addon.panel, "OptionsSliderTemplate")
knobs.showExtra = CreateFrame("CheckButton", "InsanityBarShowExtra", addon.panel, "UICheckButtonTemplate")
knobs.showTimes = CreateFrame("CheckButton", "InsanityBarShowTimes", addon.panel, "UICheckButtonTemplate")
knobs.showStacks = CreateFrame("CheckButton", "InsanityBarShowStacks", addon.panel, "UICheckButtonTemplate")
knobs.hideOutsideVF = CreateFrame("CheckButton", "InsanityBarOutsideVF", addon.panel, "UICheckButtonTemplate")
knobs.hideOutsideCombat = CreateFrame("CheckButton", "InsanityBarOutsideCombat", addon.panel, "UICheckButtonTemplate")
knobs.lock = CreateFrame("CheckButton", "InsanityBarLock", addon.panel, "UICheckButtonTemplate")
knobs.playSound = CreateFrame("CheckButton", "InsanityBarSoundPlay", addon.panel, "UICheckButtonTemplate")
knobs.apply = CreateFrame("Button", "InsanityBarApply", addon.panel, "UIPanelButtonTemplate")

-- declare variables which actually change during the addon process
local spells = {}
local extraInsFrames = {}
local vfUpThresh = 1
local swdThresh = 0
local saInFlight = 0
local maxIns = 100
local player = ""
local vfUp = nil
local vfStacks = 1
local curIns = 0
local soundIsPlaying = false
local dTime = 0


-- functions
local function AddTextToFrame(frame)
	frame.text = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	frame.text:SetPoint("RIGHT", frame, -border, 0)
	frame.text:SetTextColor(1, 1, 1)
end

local function ActivateBar()
	bgFrame:Show()
end

local function DisableBar()
	bgFrame:Hide()
end

local function InsDrain(stacks)
	-- insanity drain per second based on stacks (from voidform)
	return 6.0+(2.0/3.0)*stacks -- 7.1.5
end

local function TimeLeft(insanity, stacks)
	-- returns the remaining time until insanity would be completely drained with stacks of voidform
	-- calculation: F(x) = INTEGRAL of InsDrain(x), y = insanity-F(x+stacks)+F(stacks), get x for y = 0
	return math.sqrt(stacks*stacks+18*stacks+81+3*insanity)-stacks-9 -- 7.1.5
end

local function UpdateBar()
	-- the main routine showing current insanity and extra insanity if spells are ready
	if not (GetSpecialization() == 3) then
		-- wrong spec, disable bar completely
		DisableBar()
		return
	end
	
	if not UnitAffectingCombat("player") and InsanityBarSettings.hideOutsideCombat then
		-- user don't want to see the bar out of combat
		DisableBar()
		return
	end
	
	ActivateBar()
	
	-- current insanity level
	curIns = UnitPower("player")
	local widthNorm = barWidth/maxIns
	barFrame:SetSize(max(1, curIns)*widthNorm, barHeight)
	
	-- current drain amount
	vfUp, _, _, vfStacks = UnitBuff("player", vf)
	local timeTilDrained = 0
	if vfUp then
		-- in voidform
		drainFrame:Show()
		drainFrame:SetSize(InsDrain(vfStacks)*widthNorm, barHeight)
		if InsanityBarSettings.showTimes then
			-- get seconds until insanity would be drained to 0 for current stacks
			timeTilDrained = TimeLeft(curIns, vfStacks)
			barFrame.text:Show()
			barFrame.text:SetText(string.format("%.1f", timeTilDrained))
		else
			barFrame.text:Hide()
		end
		threshFrame:Hide()
	else
		if InsanityBarSettings.hideOutsideVF then
			-- user don't want to see the bar outside voidform
			DisableBar()
			return
		end
		drainFrame:Hide()
		barFrame.text:Hide()
		-- voidform activation threshold
		threshFrame:Show()
		threshFrame:SetPoint("CENTER", barFrame, "LEFT", barWidth*vfUpThresh, 0)
	end
	
	-- current stacks
	stackFrame.text:SetText(string.format("%d", vfStacks))
	
	-- early abort when user dont want to see extra insanity
	if not InsanityBarSettings.showExtra then
		return
	end
	
	-- insanity multiplier to all spells
	local globInsMult = 1
	-- surrender to madness
	local stmUp = UnitBuff("player", stm)
	if stmUp then
		globInsMult = globInsMult*2.0
	end
	-- power infusion
	local piUp = UnitBuff("player", pi)
	if piUp then
		globInsMult = globInsMult*1.25
	end
	
	-- append extra insanity
	local sumScaledExraIns = 0
	local prevFrame = barFrame
	for spellId, extraIns in pairs(spells) do
		-- check if spell is off cooldown
		local f = extraInsFrames[spellId]
		local start, duration = GetSpellCooldown(spellId)
		local cd = start + duration - GetTime()
		local cdReady = not (cd > 0 and duration > 1.5)
		
		-- condition for various cases
		cond = false
		if spellId == swd then
			-- special case for shadow word: death
			cond = cdReady and UnitExists("target") and UnitHealth("target")/UnitHealthMax("target") <= swdThresh
		elseif spellId == sa then
			-- special case for SA's
			cond = saInFlight > 0
			extraIns = extraIns*saInFlight
		elseif spellId == vb then
			-- special case for void bolt
			cond = cdReady and vfUp
		else
			-- normal castable spells
			cond = cdReady
		end
		
		-- show extra insanity when condition is met
		if cond then
			local scaledExtraIns = extraIns*globInsMult
			sumScaledExraIns = sumScaledExraIns+scaledExtraIns
			f:Show()
			f:SetPoint("LEFT", prevFrame, "RIGHT", 0, 0)
			f:SetSize(scaledExtraIns*widthNorm, barHeight)
			f.icon:SetSize(math.min(scaledExtraIns*widthNorm, barHeight), barHeight)
			prevFrame = f
			-- show extra time
			if vfUp and InsanityBarSettings.showTimes then
				-- get gained seconds
				local extraTimeTilDrained = TimeLeft(scaledExtraIns+curIns, vfStacks)
				f.text:Show()
				f.text:SetText(string.format("+%.1f", extraTimeTilDrained-timeTilDrained))
			else
				f.text:Hide()
			end
		else
			f:Hide()
		end
	end
	
	-- indicate capping of extra insanity
	local totalIns = curIns+sumScaledExraIns
	if totalIns > maxIns then
		capFrame:Show()
		capFrame:SetSize((totalIns-maxIns)*widthNorm, barHeight)
	else
		capFrame:Hide()
	end
end

local function CleanUpBar()
	-- aligns extra insanity left and hides them
	local prevFrame = barFrame
	for _,f in pairs(extraInsFrames) do
		f:SetPoint("LEFT", prevFrame, "RIGHT", 0, 0)
		prevFrame = f
		f:Hide()
	end
end

local function VarsToPanel()
	-- apply values to widgets
	for k, v in pairs(InsanityBarSettings) do
		if k == "width" or k == "height" then
			knobs[k]:SetValue(v)
		else
			knobs[k]:SetChecked(v)
		end
	end
end

local function PanelToVars()
	-- get settings from widget values
	for k, _ in pairs(InsanityBarSettings) do
		if k == "width" or k == "height" then
			InsanityBarSettings[k] = knobs[k]:GetValue()
		else
			InsanityBarSettings[k] = knobs[k]:GetChecked()
		end
	end
	
	-- apply some values directly for direct response
	-- show/hide bar
	if not InsanityBarSettings.hideOutsideVF then
		-- bar needs to be updated regularly because there is no event that fires when spells come off cooldown
		bgFrame:SetScript("OnUpdate", function(self, elapsed)
			dTime = dTime+elapsed
			if dTime > 0.5 then
				UpdateBar()
				dTime = 0
			end
		end)
	else
		DisableBar()
		bgFrame:SetScript("OnUpdate", nil)
	end
	-- lock/unlock
	if InsanityBarSettings.lock then
		bgFrame:EnableMouse(false)
	else
		bgFrame:EnableMouse(true)
	end
	-- hide extra insanity
	if not InsanityBarSettings.showExtra then
		CleanUpBar()
	end
	-- size
	barWidth = InsanityBarSettings.width
	barHeight = InsanityBarSettings.height
	-- background
	bgFrame:SetSize(barWidth+2*border, barHeight+2*border)
	threshFrame:SetSize(2, barHeight+2*border)
	-- stack frame
	if InsanityBarSettings.showStacks then
		stackFrame:Show()
		stackFrame:SetSize(barHeight+2*border, barHeight+2*border)
	else
		stackFrame:Hide()
	end
	
	UpdateBar()
end

local function ResetSettings()
	-- reset settings which are saved between game sessions
	-- global saved variables
	InsanityBarSettings = {}
	for k, v in pairs(defaultSettings) do
		InsanityBarSettings[k] = v
	end
	print('InsanityBar settings reset')
	VarsToPanel()
end

local function CheckSettings()
	-- check if settings exists
	if InsanityBarSettings == nil then
		ResetSettings()
	end
	-- check if keys from defaults exists in settings
	for k, v in pairs(defaultSettings) do
		if InsanityBarSettings[k] == nil then
			print('InsanityBar got new setting: '..k)
			InsanityBarSettings[k] = v
		end
	end
	-- check if keys from settings exists in defaults
	for k, v in pairs(InsanityBarSettings) do
		if defaultSettings[k] == nil then
			print('InsanityBar removed deprecated setting: '..k)
			InsanityBarSettings[k] = nil
		end
	end
	-- update GUI
	VarsToPanel()
end

local function BuildSettingsPanel()
	-- only a convienience thing for the user
	-- options panel (esc -> Interface -> Addons)
	
	-- Add the panel to the Interface Options
	InterfaceOptions_AddCategory(addon.panel);
	local vSpace = 30 -- spacing between elements
	local hPos, vPos = 20, -20 -- start position of elements
	
	-- build slider for bar width
	local f = knobs.width
	f:SetPoint("TOPLEFT", hPos+5, vPos)
	f:SetSize(500, 20)
	_G[f:GetName()..'Text']:SetText('Bar width') -- sets the "title" text (top-centre of slider).
	_G[f:GetName()..'Low']:SetText('1') -- sets the left-side slider text (default is "Low").
	_G[f:GetName()..'High']:SetText('1500') -- sets the right-side slider text (default is "High").
	f:SetMinMaxValues(1, 1500)
	f:SetValueStep(1)
	f:SetObeyStepOnDrag(true) -- actually enable step behaviour
	f:SetScript("OnValueChanged", function(self, value)
		_G[self:GetName()..'Text']:SetText('Bar width: '..self:GetValue())
	end)
	vPos = vPos - vSpace
	
	-- build slider for bar height
	f = knobs.height
	f:SetPoint("TOPLEFT", hPos+5, vPos)
	f:SetSize(500, 20)
	_G[f:GetName()..'Text']:SetText('Bar height')
	_G[f:GetName()..'Low']:SetText('1')
	_G[f:GetName()..'High']:SetText('150')
	f:SetMinMaxValues(1, 100)
	f:SetValueStep(1)
	f:SetObeyStepOnDrag(true)
	f:SetScript("OnValueChanged", function(self, value)
		_G[self:GetName()..'Text']:SetText('Bar height: '..self:GetValue())
	end)
	vPos = vPos - vSpace
	
	-- build checkbox to show or hide extra insanity
	f = knobs.showExtra
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Show extra insanity from spells when available")
	vPos = vPos - vSpace
	
	-- build checkbox to show or hide drain numbers
	f = knobs.showTimes
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Show drain times")
	vPos = vPos - vSpace
	
	-- build checkbox to show or hide stacks
	f = knobs.showStacks
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Show insanity stacks")
	vPos = vPos - vSpace

	-- build checkbox to hide bar outside voidform
	f = knobs.hideOutsideVF
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Hide bar outside voidform")
	vPos = vPos - vSpace
	
	-- build checkbox to hide bar outside combat
	f = knobs.hideOutsideCombat
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Hide bar outside combat")
	vPos = vPos - vSpace
	
	-- build checkbox to lock bar
	f = knobs.lock
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Lock bar")
	vPos = vPos - vSpace
	
	-- build checkbox to play a sound when insanity reaches threshold
	f = knobs.playSound
	f:SetPoint("TOPLEFT", hPos, vPos)
	_G[f:GetName()..'Text']:SetText("Play sound when voidform can be entered")
	vPos = vPos - vSpace
	
	-- build apply button
	f = knobs.apply
	f:SetPoint("BOTTOMLEFT", hPos, 20)
	f:SetSize(100, 20)
	f:SetText("Apply")
	f:SetScript("OnClick", function(self)
		PanelToVars()
	end)
	
	-- load setting values from saved variabled and apply to GUI
	VarsToPanel()

	-- function called when pressing the "default" button
	addon.panel.default = function(self)
		ResetSettings()
	end

	-- function called when pressing the "okay" button
	addon.panel.okay = function(self)
		PanelToVars()
	end
end

local function BuildBar()
	-- building the static frames for the insanity bar
	
	-- build background frame
	bgFrame:SetPoint("CENTER", UIParent)
	bgFrame.tex = bgFrame:CreateTexture()
	bgFrame.tex:SetAllPoints(bgFrame)
	bgFrame.tex:SetColorTexture(0.1, 0.1, 0.1, 0.5)
	-- make moveable by mouse
	bgFrame:SetMovable(true)
	bgFrame:RegisterForDrag("LeftButton")
	bgFrame:SetScript("OnDragStart", bgFrame.StartMoving)
	bgFrame:SetScript("OnDragStop", bgFrame.StopMovingOrSizing)
	bgFrame:SetUserPlaced(true)
	
	-- build bar frame
	barFrame:SetPoint("LEFT", bgFrame, "LEFT", border, 0)
	barFrame.tex = barFrame:CreateTexture()
	barFrame.tex:SetAllPoints(barFrame)
	barFrame.tex:SetColorTexture(0.4, 0.2, 1.0, 0.7)
	AddTextToFrame(barFrame)
	
	-- build stack frame
	stackFrame:SetPoint("RIGHT", bgFrame, "LEFT", -border, 0)
	stackFrame.tex = stackFrame:CreateTexture()
	stackFrame.tex:SetAllPoints(stackFrame)
	stackFrame.tex:SetColorTexture(0.1, 0.1, 0.1, 0.5)
	AddTextToFrame(stackFrame)
	
	-- build drain frame
	drainFrame:SetPoint("LEFT", bgFrame, "LEFT", border, 0)
	drainFrame.tex = drainFrame:CreateTexture()
	drainFrame.tex:SetAllPoints(drainFrame)
	drainFrame.tex:SetColorTexture(0.5, 0.0, 0.0, 0.5)
	drainFrame:SetFrameStrata("DIALOG")
	
	-- build cap indicator
	capFrame:SetPoint("LEFT", bgFrame, "RIGHT", -border, 0)
	capFrame.tex = capFrame:CreateTexture()
	capFrame.tex:SetAllPoints(capFrame)
	capFrame.tex:SetColorTexture(0.2, 0.2, 0.2, 0.5)
	capFrame:SetFrameStrata("DIALOG")
	
	-- build threshold indicator
	threshFrame.tex = threshFrame:CreateTexture()
	threshFrame.tex:SetAllPoints(threshFrame)
	threshFrame.tex:SetColorTexture(1, 1, 1, 0.5)
end

local function UpdateInsanitySpells()
	-- dictionary of spells that grant a good amount of insanity and their corresponding amount
	-- the DoTs are not included here, but spells that could prevent dropping out of shadow form

	-- check for RoS multiplier
	swdThresh = 0.2
	local _, _, _, selected = GetTalentInfo(4, 2, 1)
	if selected then
		swdThresh = 0.35
	end
	
	-- check for insanity activation threshold
	vfUpThresh = 1
	local _, _, _, selected = GetTalentInfo(7, 1, 1)
	if selected then
		vfUpThresh = 0.65
	end
	
	spells = {}
	spells[mb] = 12 -- mindblast
	spells[vb] = 16 -- void bolt
	spells[swd] = 15 -- shadow word: death
	
	-- check for talents selected that grant insanity
	local _, _, _, selected = GetTalentInfo(1, 3, 1)
	if selected then
		spells[swv] = 25 -- shadow word: void
	end
	local _, _, _, selected = GetTalentInfo(5, 2, 1)
	if selected then
		spells[sa] = 3 -- shadowy appariations
		bgFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		bgFrame:RegisterEvent("UNIT_SPELLCAST_FAILED_QUIET")
	else
		bgFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		bgFrame:UnregisterEvent("UNIT_SPELLCAST_FAILED_QUIET")
		saInFlight = 0
	end
	local _, _, _, selected = GetTalentInfo(7, 2, 1)
	if selected then
		spells[sc] = 15 -- shadow crash
	end
	
	-- build extra insanity frames
	for spellId, extraIns in pairs(spells) do
		if not extraInsFrames[spellId] then
			-- make new frame because not yet in list
			local f = CreateFrame("Frame", "ExtraInsanityBarFrame"..spellId, barFrame)
			f:SetPoint("LEFT", barFrame, "RIGHT", 0, 0)
			f.tex = f:CreateTexture("ExtraInsanityBarTex"..spellId)
			f.tex:SetAllPoints(f)
			f.tex:SetColorTexture(0.7, 0.9, 1.0, 0.7)
			-- add text frame
			AddTextToFrame(f)
			
			-- also make icon
			local g = CreateFrame("Frame", "ExtraInsanityIconFrame"..spellId, f)
			g:SetPoint("LEFT", f, "LEFT", 0, 0)
			g.tex = g:CreateTexture("ExtraInsanityIconTex"..spellId)
			g.tex:SetAllPoints(g)
			local texFile = GetSpellTexture(spellId)
			g.tex:SetTexture(texFile)
			
			-- link and hide
			f.icon = g
			f:Hide()
			-- add to list
			extraInsFrames[spellId] = f
		end
	end
end

-- watchdog to reset SA counter in case something goes wrong
local saWatchdog = 0
mainFrame:SetScript("OnUpdate", function(self, elapsed)
	if (saWatchdog > 0) then
		saWatchdog = saWatchdog-elapsed
	else
		saInFlight = 0
	end
end)


-- events
mainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
mainFrame:RegisterEvent("PLAYER_TALENT_UPDATE")
mainFrame:RegisterEvent("ADDON_LOADED")
mainFrame:RegisterEvent("UNIT_AURA")
mainFrame:SetScript("OnEvent", function(self, event, ...)
	-- load settings
	if event == "ADDON_LOADED" then
		if not addon.loaded then
			addon.loaded = true
			CheckSettings()
			BuildSettingsPanel()
			BuildBar()
		end
	end
	
	-- look for talents selected
	if event == "PLAYER_ENTERING_WORLD" or event == "PLAYER_TALENT_UPDATE" then
		player = UnitName("player")
		CleanUpBar()
		UpdateInsanitySpells()
		PanelToVars()
		UpdateBar()
	end
	
	-- look if voidform is up
	if event == "UNIT_AURA" then
		local unitId = ...
		if unitId == "player" then
			UpdateBar()
		end
	end
end)

bgFrame:RegisterEvent("UNIT_POWER_FREQUENT")
bgFrame:SetScript("OnEvent", function(self, event, ...)
	-- this is fired when the main resource of a unit changes
	if event == "UNIT_POWER_FREQUENT" then
		local unit, powerType = ...
		if unit == "player" and powerType == "INSANITY" then
			UpdateBar()
			-- play sound when voidform threshold is reached
			if InsanityBarSettings.playSound then
				if curIns >= vfUpThresh*maxIns and not vfUp and not soundIsPlaying then
					PlaySoundFile(soundPath)
					soundIsPlaying = true
				end
				if curIns < vfUpThresh*maxIns then
					soundIsPlaying = false
				end
			end
		end
	end
	
	-- this is fired for every combat action
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		-- we need this to track how many SA's are on the way (ideally)
		local _, spellEvent, _, _, sourceUnit, _, _, _, _, _, _, spellId, _, _, _, _, _, _, _, _, critical = ...
		-- only track stuff from us
		if sourceUnit == player then
			-- track shadowy appariations impact
			if spellId == sa and spellEvent == "SPELL_DAMAGE" then
				saInFlight = saInFlight-1
			end
			-- track crits from sw:p
			if spellId == swp and spellEvent == "SPELL_PERIODIC_DAMAGE" and critical then
				saWatchdog = 10 -- reset watchdog to x seconds
				saInFlight = saInFlight+1
			end
		end
	end
	
	-- fired for unusual spell fails
	if event == "UNIT_SPELLCAST_FAILED_QUIET" then
		-- we need this to track SA's which can't reach the target (because it is dead or so)
		local sourceUnit, _, _, _, spellId = ...
		if sourceUnit == player and spellId == sa then
			saInFlight = saInFlight-1
		end
	end
end)
